#!/usr/bin/env python
#
# GABI - Genome Assembly by Bayesian Inference
#
# Copyright 2013-2014, Brown University, Providence, RI. All Rights Reserved.
#
# This file is part of GABI.
# 
# GABI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# GABI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with GABI.  If not, see <http://www.gnu.org/licenses/>.

import math
import matplotlib
matplotlib.use('Agg')

import numpy as np
import os
import sqlite3
import sys

from itertools import chain, combinations
from operator import itemgetter
from matplotlib.pyplot import *

matplotlib.rcParams.update({'font.size': 8})

palette = ('r', 'b', 'g')
palette2 = ('m', 'y', 'c')

def convergence_plot(traces, outname):
	"""
	Generates:
	(a) Mixing behvaior
	(b) Cumulative frequencies of individual edges in each chain
	(c) Average st. dev. of the cumulative edge frequencies across chains
	(d) Split frequencies at the final sample
	"""

	likelihoods = []
	nsamples = []
	edges = []
	frequencies = []

	for trace in traces:
		db = sqlite3.connect(trace)
		sql = """
			SELECT likelihood 
			FROM samples
			WHERE accept=1
			ORDER BY id;"""
		likelihoods.append(map(itemgetter(0), db.execute(sql)))
		nsamples.append(len(likelihoods[-1]))
		sql = """
			SELECT edge
			FROM states JOIN samples ON states.id=samples.id
			WHERE accept=1
			GROUP BY edge;"""
		edges.append([e[0] for e in db.execute(sql)])
		frequencies.append(dict([(e, np.zeros(nsamples[-1])) for e in edges[-1]]))
		sql = """
			SELECT edge, states.id
			FROM states JOIN samples ON states.id=samples.id
			WHERE samples.accept=1
			ORDER by states.id;"""
		i = 0
		last_id = 0
		for e, id in db.execute(sql): 
			frequencies[-1][e][i] = 1.0
			if last_id != id:
				i += 1
				last_id = id
		db.close()

	figure(figsize=(8,6))

	subplot(221)
	title('a', size=12, weight='heavy')
	xlabel('Accepted Samples')
	ylabel('log(likelihood)')
	xmax = max(nsamples)
	xlim([-0.02*xmax, 1.02*xmax])
	grid(axis='x')
	gca().set_rasterization_zorder(1)
	for i, y in enumerate(likelihoods):
		label = 'Chain %d' % (i+1)
		plot(y, color=palette[i%3], alpha=0.5, zorder=0, label=label)
	gca().legend(loc='lower right', fontsize=7)

	# Calculate cumulative frequencies
	for i, f in enumerate(frequencies):
		for e in f:
			f[e] = np.cumsum(f[e]) / np.arange(1, nsamples[i]+1)

	# Establish edge order
	edges = sorted(set(chain(*edges)))

	subplot(222)
	title('b', size=12, weight='heavy')
	xlabel('Accepted Samples')
	ylabel('Edge Frequency')
	xlim([-0.02*xmax, 1.02*xmax])
	grid(axis='x')
	gca().set_rasterization_zorder(1)
	for i, f in enumerate(frequencies):
		for e in f:
			plot(f[e], color=palette[i%3], alpha=0.1, zorder=0)

	# Average stdev of the split frequencies
	xmin = min(nsamples)
	stdev = np.zeros(xmin)
	for i in xrange(xmin):
		stdev[i] = sum(np.std([f.get(e, np.zeros(i+1))[i] for f in frequencies]) for e in edges)
	stdev *= 1.0 / len(edges)

	subplot(223)
	title('c', size=12, weight='heavy')
	xlabel('Accepted Samples')
	ylabel('Average St. Dev.')
	xlim([-0.02*xmin, 1.02*xmin])
	grid(axis='x')
	plot(stdev, color='k')

	subplot(224)
	title('d', size=12, weight='heavy')
	xlim([0, 1])
	ylim([0, 1])
	if len(frequencies) == 2:
		xlabel('Edge Frequency (Chain 1)')
		ylabel('Edge Frequency (Chain 2)')
		x = [frequencies[0].get(e, [0.0])[-1] for e in edges]
		y = [frequencies[1].get(e, [0.0])[-1] for e in edges]
		scatter(x, y, facecolors='none', edgecolors='k')
	else:
		xlabel('Edge Frequency (Chain X)')
		ylabel('Edge Frequency (Chain Y)')
		for i, pair in enumerate(combinations(xrange(len(frequencies)), 2)):
			a, b = pair
			x = [frequencies[a].get(e, [0.0])[-1] for e in edges]
			y = [frequencies[b].get(e, [0.0])[-1] for e in edges]
			scatter(
				x, y, label='Chain %d vs. %d'%(a+1, b+1),
				facecolors='none', edgecolors=palette2[i])
		gca().legend(loc='lower right', fontsize=7)

	tight_layout()
	savefig(outname, dpi=300)


if __name__ == '__main__':
	convergence_plot(sys.argv[1:], 'convergence.pdf')

