#!/usr/bin/env python
#
# GABI - Genome Assembly by Bayesian Inference
#
# Copyright 2013-2014, Brown University, Providence, RI. All Rights Reserved.
#
# This file is part of GABI.
#
# GABI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# GABI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with GABI.  If not, see <http://www.gnu.org/licenses/>.

import math
import networkx as nx
import os
import sys

from random import choice, randint

from Bio import SeqIO
from Bio.Alphabet import DNAAlphabet
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from biolite import utils

# Graph class
class Graph:
	def __init__(self, graphml):
		self.G = nx.read_graphml(graphml, node_type=int)
		self.edges = [None] * self.G.number_of_edges()
		for n1, n2, k in self.G.edges_iter(keys=True):
			self.edges[self.G[n1][n2][k]['e']-1] = (n1, n2, k)
		self.state = [False] * len(self.edges)
		self.A = nx.DiGraph() # assembly graph
		self.A.add_nodes_from(self.G.nodes_iter())

	# Graph accessors

	def len(self, n):
		edges = self.G.out_edges(n, keys=True)
		return 1.0 * sum(len(e[2]) for e in edges) / len(edges)
	def indeg(self, n):
		deg = self.A.in_degree(n)
		assert deg <= 1
		return deg
	def outdeg(self, n):
		deg = self.A.out_degree(n)
		assert deg <= 1
		return deg
	def inedge(self, n):
		edges = self.A.in_edges(n, data=True)
		assert len(edges) <= 1
		return edges[0][2]['e']
	def outedge(self, n):
		edges = self.A.out_edges(n, data=True)
		assert len(edges) <= 1
		return edges[0][2]['e']
	def inedge_random(self, n):
		edges = self.G.in_edges(n, data=True)
		return choice(edges)[2]['e']
	def outedge_random(self, n):
		edges = self.G.out_edges(n, data=True)
		return choice(edges)[2]['e']
	def isactivenode(self, n):
		return self.indeg(n) + self.outdeg(n)

	def seq(self, e): return self.edges[e-1][2]
	def head(self, e): return self.edges[e-1][1]
	def tail(self, e): return self.edges[e-1][0]
	def isactive(self, e): return self.state[e-1]

	# Graph mutators

	def activate(self, e):
		n1, n2, k = self.edges[e-1]
		self.A.add_edge(n1, n2, e=e)
		self.state[e-1] = True

	def deactivate(self, e):
		n1, n2, k = self.edges[e-1]
		if self.A.has_edge(n1, n2):
			self.A.remove_edge(n1, n2)
		self.state[e-1] = False

	# Graph output

	def contig_header(self, path):
		header = []
		for e in path:
			header.append('#node%d' % self.tail(e))
			header.append('#edge%d' % e)
		return ','.join(header)

	def contig_record(self, path, i, desc=""):
		"""
		Build a FASTA SeqRecord for the `path` through the graph.
		"""
		if self.tail(path[0]) == self.head(path[-1]):
			# Circular
			seq = ''.join(map(self.seq, path))
		else:
			# Linear
			seq = self.G.node[self.head(path[0])]['kmer'] + ''.join(map(self.seq, path))
		return SeqRecord(Seq(seq), 'contig-%d' % i, description=str(len(seq))+desc)

	def contigs(self, records=True, paths=False):
		"""
		An iterator that yields a FASTA SeqRecord for each path (linear or
		cicular) through the current state.
		"""
		ncontigs = 0
		edges = set(e[2]['e'] for e in self.A.edges_iter(data=True))
		# First, find all linear contigs starting at a node with an out-edge
		# but no in-edge.
		for n in self.A.nodes_iter():
			if self.outdeg(n) and not self.indeg(n):
				path = []
				while self.outdeg(n):
					e = self.outedge(n)
					n = self.head(e)
					path.append(e)
					edges.add(e) # have to add first, to remove w/out except
					edges.remove(e)
				if records:
					yield self.contig_record(path, ncontigs)
				elif paths:
					yield path
				else:
					yield self.contig_header(path)
				ncontigs += 1
		# Next, find all circular contigs from the remaining edges.
		while edges:
			e = edges.pop()
			path = [e]
			start = self.tail(e)
			while self.head(e) != start:
				e = self.outedge(self.head(e))
				path.append(e)
				edges.add(e) # have to add first, to remove w/out except
				edges.remove(e)
			if records:
				yield self.contig_record(path, ncontigs)
			elif paths:
				yield path
			else:
				yield self.contig_header(path)
			ncontigs += 1

	def fasta_edges(self):
		for e, edge in enumerate(self.edges, start=1):
			n1, n2, seq = edge
			seq = self.G.node[n1]['kmer'] + seq
			yield SeqRecord(Seq(seq), 'edge%d' % e, description=str(len(seq)))

	def contigs_annotated(self, posteriors):
		# Reset assembly graph.
		self.state = [False] * len(self.edges)
		self.A = nx.DiGraph()
		self.A.add_nodes_from(self.G.nodes_iter())
		for e in posteriors:
			self.activate(e)
		for i, path in enumerate(self.contigs(records=False, paths=True)):
			p = []
			l = len(self.G.node[self.head(path[0])]['kmer'])
			for e in path:
				n1, n2, seq = self.edges[e-1]
				p.append('%d-%d: %f' % (l, l+len(seq)-1, posteriors[e]))
				l += len(seq)
			yield self.contig_record(
				path, i, desc=" posteriors=[%s]" % ', '.join(p))

	def write_dot(self, filename, frequencies=None, colors=None, active=False):
		"""
		Write a dot file for the graph, highlighting the selected nodes and
		edges.
		"""
		norm = 1.0 / max(map(math.sqrt, map(self.len, self.G.nodes_iter())))
		palette = (
			(253, 212, 158),
			(253, 187, 132),
			(252, 141, 89),
			(239, 101, 72),
			(215, 48, 31),
			(179, 0, 0),
			(127, 0, 0))
		with open(filename, 'w') as f:	
			print >>f, 'digraph G {'
			for n in sorted(self.G.nodes()):
				width = math.sqrt(self.len(n)) * norm
				print >>f, '"%d" [id="node%d" label="" shape=circle color=grey width=%f];' % (n, n, width)
			for e, edge in enumerate(self.edges, start=1):
				if frequencies:
					if e in frequencies:
						x = frequencies[e] / 0.16
						i = int(x)
						if i < 6:
							x -= i
							assert x < 1.0
							c = (
								x * palette[i][0] + (1.0 - x) * palette[i+1][0],
								x * palette[i][1] + (1.0 - x) * palette[i+1][1],
								x * palette[i][2] + (1.0 - x) * palette[i+1][2])
						else:
							c = palette[i]
						c = '#%02x%02x%02x' % c
						w = 5 * frequencies[e]
						#style = 'color="%s" arrowsize=%f penwidth=%f' % (c, frequencies[e], w)
						style = 'color="%s" arrowsize=%f penwidth=4.0' % (c, frequencies[e])
					else:
						#style = "color=grey arrowsize=0.5 penwidth=0.5 style=dashed"
						style = "color=grey90 penwidth=4.0"
				elif colors:
					if e in colors:
						style = 'color="%s" penwidth=4.0' % colors[e]
					else:
						style = "color=grey90 penwidth=4.0"
				elif active:
					if self.state[e-1]:
						style = "color=red penwidth=4.0"
					else:
						style = "color=grey90 penwidth=4.0"
				else:
					style = "color=grey arrowsize=0.5"
				#weight = math.exp(self.G.out_degree(edge[0]) + self.G.in_degree(edge[1]))
				weight = 1.0
				print >>f, '"%d" -> "%d" [id="edge%d" weight=%f %s minlen=10000];' % (edge[0], edge[1], e, weight, style)
			print >>f, '}'

	# Graph state

	def set_state(self, state):
		self.state = list(state)
		self.A = nx.DiGraph() # assembly
		self.A.add_nodes_from(self.G)
		map(self.activate, (e for e, on in enumerate(self.state, start=1) if on))

	def get_state(self):
		return self.state

	def copy_state(self):
		return list(self.state)

	def hash_state(self):
		return ''.join(map(lambda x : str(x)[0], self.state))

	def perturb(self):
		chosen = randint(1, len(self.edges))
		if self.isactive(chosen):
			self.deactivate(chosen)
		else:
			path = [chosen]
			path_nodes = set()
			path_nodes.add(self.tail(chosen))
			# Search forward
			n = self.head(chosen)
			while not (self.isactivenode(n) or n in path_nodes):
				e = self.outedge_random(n)
				path.append(e)
				path_nodes.add(n)
				n = self.head(e)
			end = n
			# Search backward
			n = self.tail(chosen)
			while not (self.isactivenode(n) or n in path_nodes):
				e = self.inedge_random(n)
				path.insert(0, e)
				path_nodes.add(n)
				n = self.tail(e)
			start = n
			# Deactivate old path
			if self.outdeg(start):
				e = self.outedge(start)
				n = self.head(e)
				while n != end and self.outdeg(n):
					self.deactivate(e)
					e = self.outedge(n)
					n = self.head(e)
				self.deactivate(e)
			if self.indeg(end):
				e = self.inedge(end)
				n = self.tail(e)
				while n != start and self.indeg(n):
					self.deactivate(e)
					e = self.inedge(n)
					n = self.tail(e)
				self.deactivate(e)
			# Activate new path
			map(self.activate, path)
		return chosen

	def randomize_state(self):
		#self.perturb()
		for i in xrange(len(self.state)):
			chosen = randint(1, len(self.edges))
			n0, n1, seq = self.edges[chosen-1]
			if self.outdeg(n0) == 0 and self.indeg(n1) == 0:
				self.activate(chosen)
		self.write_dot('random.dot', active=True)

