#!/usr/bin/env python
#
# GABI - Genome Assembly by Bayesian Inference
#
# Copyright 2013-2014, Brown University, Providence, RI. All Rights Reserved.
#
# This file is part of GABI.
#
# GABI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GABI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GABI.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import os
import sqlite3
import subprocess

from convergence import convergence_plot
from consensus import consensus_plots, posterior_plot, majority_rule_plot
from graph import Graph

from biolite import diagnostics
from biolite import report
from biolite import utils

html_template = """<!DOCTYPE html>
<head>
<meta http-equiv="content-type" content="text/html;charset=utf-8">
<title>%s</title>
<link href="css/bootstrap.min.css" rel="stylesheet" media="screen">
<style type="text/css">
body {
  margin: 30px 50px;
}
.stats {
  margin: 10px;
  display: none;
}
.table {
  font-size: 14px;
  font-family: Monaco,Menlo,Consolas,"Courier New",monospace;
}
.table-mini {
  width: auto;
}
table.layout tr {
  border-top: 1px solid #eee;
  border-bottom: 1px solid #eee;
}
blockquote p { font-size: 14px; }
.axis path, .axis line {
  fill: none;
  stroke: black;
  shape-rendering: crispEdges;
}
.axis text {
  font-family: sans-serif;
  font-size: 10px;
}
</style>
</head>
<body>
%s
</body>
</html>
"""

animation_template = """
<h2 id="trace-header">Loading...</h2>
<h5 style="text-align:center">log(likelihood)</h5>
<div id="trace"></div>
<p style="text-align:center" id="step"></p>
<blockquote>Mouse over the plot to see the assembly of a particular sample.</blockquote>
</div>
<div id="assembly" style="width:100%%; float:left"></div>
<script type="text/javascript" src="js/jquery.min.js"></script>
<script type="text/javascript" src="js/d3.min.js"></script>
<script type="text/javascript" src="%s"></script>
<script type="text/javascript">
var assembly = d3.select("#assembly");

// Colors
var grey="silver";
var palette = [
  "#FB8072",
  "#BEBADA",
  "#8DD3C7",
  "#FFED6F",
  "#80B1D3",
  "#FDB462",
  "#B3DE69",
  "#FCCDE5",
  "#BC80BD",
  "#CCEBC5" ];

// Animation state

var i = 0;
var end = likelihoods.length - 1;
var playing = false;

// Trace plot

var height = 200;
var width = $("#summary-table").width() - 30;
var trace = d3.select("#trace").append("svg:svg").attr("width", width+30).attr("height", height+30);
var x;
var y;

function clear_contigs() {
  assembly.selectAll(".node").selectAll("ellipse").attr("stroke", grey).attr("fill-opacity", "0.0");
  assembly.selectAll(".edge").selectAll("path,polygon").attr("stroke", grey);
}

function color_contigs() {
  for (j=0; j<contigs[i].length; j++) {
    var color = palette[j %% palette.length];
    assembly.selectAll(contigs[i][j]).selectAll("path,polygon,ellipse").attr("stroke", color);
    assembly.selectAll(contigs[i][j]).selectAll("ellipse").attr("fill", color).attr("fill-opacity", "0.3");
  }
}

function redraw() {
  d3.select("#step").text("Sample #:  "+i);
  trace.select("#marker").attr("d", "M"+(x(i)+30)+",0V"+height);
  clear_contigs();
  color_contigs();
}

function step() {
  redraw();
  i = (i + 1) %% end;
}

var interval;
function animate() {
  if (playing) {
    playing = false;
    interval = window.clearInterval(interval);
  } else {
    playing = true;
    interval = setInterval(step, 34);
  }
}

function trace_hover() {
  if (!playing) {
    j = Math.round(x.invert(d3.mouse(this)[0]-30));
    if (i != j && j >= 0) {
      i = j;
      redraw();
    }
  }
}

// Trace plot

x = d3.scale.linear().domain([0, likelihoods.length - 1]).range([0,width]);
y = d3.scale.linear().domain([d3.min(likelihoods), d3.max(likelihoods)]).range([height,0]);
trace.selectAll("path.line").data([likelihoods]).enter().append("svg:path").attr("d", d3.svg.line().x(function(d,i) { return  x(i); }).y(function(d) { return y(d); })).attr("stroke", "blue").attr("stroke-width", 0.5).attr("fill", "none").attr("transform", "translate(30,0)");
trace.append("path").attr("id", "marker").attr("d", "M30,0V"+height).attr("stroke", 'red');
var xAxis = d3.svg.axis().scale(x).orient("bottom");
var yAxis = d3.svg.axis().scale(y).orient("left");
trace.append("g").attr("class","axis").attr("transform", "translate(30," + height + ")").call(xAxis);
trace.append("g").attr("class","axis").attr("transform", "translate(30,0)").call
(yAxis);

// Assembly graph

function appendStringAsNodes(id, xml) {
    var frag = document.createDocumentFragment(),
        tmp = document.createElement('body'), child;
    tmp.innerHTML = xml;
    // Append elements in a loop to a DocumentFragment, so that the browser does
    // not re-render the document for each node
    while (child = tmp.firstChild) {
        frag.appendChild(child);
    }
    d3.select(id).node().appendChild(frag); // Now, append all elements at once
    frag = tmp = null;
}
appendStringAsNodes("#assembly", assembly_svg);

assembly_svg = assembly.select("svg");
var aspect = parseInt(assembly_svg.attr("width")) / parseInt(assembly_svg.attr("height"));
width = 0.99 * ($("#assembly").width() - width - 100);
var height = aspect*width;
$("#assembly").width(width).height(height);
assembly_svg.attr("width", width).attr("height", height);

clear_contigs();
color_contigs();

// Mouse events

assembly.on("click", animate);
trace.on("click", animate);
trace.on("mousemove", trace_hover);

d3.select("#trace-header").text("Trace Animation");
d3.select("#step").text("Sample #:  0");

</script>
"""


def lookup_args(run_ids):
	"""
	Lookup the runs in the diagostics database, and associated paths and
	arguments
	"""
	graphml = None
	args = {}
	for run_id in sorted(run_ids):
		run = diagnostics.lookup_run(run_id)
		if run.name != 'sample':
			utils.die("run", run_id, "is not a 'sample' run")
		init = diagnostics.lookup(run_id, diagnostics.INIT)
		if not init or 'graphml' not in init:
			utils.die("could not load diagnostics for run", run_id)
		if graphml is None:
			graphml = init['graphml']
		elif graphml != init['graphml']:
			utils.die("run", run_id, "was based on a different graph")
		trace = diagnostics.lookup(run_id, 'sample.sample.trace')
		if not 'path' in trace:
			utils.die("couldn't find trace path for run", run_id)
		init['run'] = run
		init['trace'] = trace['path']
		args[run_id] = init
	return args


def runtime_stats(db, args):
	sql = """
		SELECT COUNT(runtime), MIN(runtime), MAX(runtime), SUM(runtime)
		FROM samples;"""
	nsamples, mintime, maxtime, time = map(float, db.execute(sql).fetchone())
	sql = "SELECT COUNT(*) FROM samples WHERE accept=1;"
	accept = int(db.execute(sql).fetchone()[0])
	args['nsamples'] = nsamples
	args['mintime'] = mintime
	args['maxtime'] = maxtime
	args['time'] = time
	args['accept'] = accept
	args['accept_per'] = 100.0 * accept * int(args['thinning']) / nsamples


def trace_animation(run_id, db, graph, svg, outdir):

	ids = []
	likelihoods = []

	sql = "SELECT id, likelihood FROM samples WHERE accept=1 ORDER BY id;"
	for id, likelihood in db.execute(sql):
		ids.append(id)
		likelihoods.append(str(likelihood))

	state = graph.copy_state()
	contigs = []

	for i, id in enumerate(ids):
		state = [False] * len(state)
		sql = "SELECT edge FROM states WHERE id=?;"
		for edge in db.execute(sql, (id,)):
			state[int(edge[0])-1] = True
		graph.set_state(state)
		contigs.append('["%s"]' % '","'.join(sorted(
			graph.contigs(records=False),
			key=(lambda s : s.count(',')),
			reverse=True)))

	data_js = 'run%d.data.js' % run_id
	with open(os.path.join(outdir, data_js), 'w') as f:
		print >>f, "var assembly_svg = '", svg, "';"
		print >>f, "var likelihoods = [", ','.join(likelihoods), "];"
		print >>f, "var contigs = [", ','.join(contigs), "];"

	return animation_template % data_js


def run_report(run_id, args, graph, svg, outdir):

	db = sqlite3.connect(args['trace'])

	runtime_stats(db, args)
	table = [
		('Catalog ID', args['run'].id),
		('Timestamp', args['run'].timestamp),
		('Expected # Contigs', args['contigs']),
		('Expected Genome Size', '{:,} bp'.format(int(args['genome_size']))),
		('Thinning', args['thinning']),
		('# Samples', '{:,.0f}'.format(args['nsamples'])),
		('# Accepted', '{:,} ({:.1f}%)'.format(args['accept'], args['accept_per'])),
		('Total Sample Time', '{:,.1f} sec'.format(args['time'])),
		('Min. Sample Time', '%.1f sec' % args['mintime']),
		('Max. Sample Time', '%.1f sec' % args['maxtime']),
		('Mean Sample Time', '%.1f sec' % (args['time'] / args['nsamples']))]
	row = '<tr><td><b>{}</b></td><td>{}</td></tr>'
	table = '\n'.join(row.format(k, v) for k, v in table)

	title = 'GABI Run %d' % run_id
	body = [
		'<div id="wrapper" style="float:left">',
		'<h1><em>%s</em></h1>' % title,
		'<blockquote>',
		'<table id="summary-table" class="table table-striped table-condensed table-mini">',
		table,
		'</table></blockquote>',
		trace_animation(run_id, db, graph, svg, outdir)]

	html = html_template % (title, '\n'.join(body))
	open(os.path.join(outdir, 'run%d.html' % run_id), 'w').write(html)

	args['frequencies'] = consensus_plots(graph, db, 0.5, os.path.join(outdir, 'run%d' % run_id))

	db.close()

	return table


def index_report(run_ids, outdir):
	args = lookup_args(run_ids)

	graph = Graph(next(args.itervalues())['graphml'])
	graph_dot = os.path.join(outdir, 'graph.dot')
	graph.write_dot(graph_dot)
	svg = subprocess.check_output(['dot', '-Tsvg', '-Kneato', graph_dot])
	svg = svg.replace('\n','')

	title = "GABI Report (Runs %s)" % ','.join(utils.number_range(run_ids))
	body = [
		'<h1><em>%s</em></h1>' % title,
		'<blockquote>',
		'Summary report for multiple GABI runs. Click on a Run header',
		'below for a detailed interactive trace for that run.',
		'</blockquote>',
		'<table class="layout">']

	# Stats and traces for each run.
	for run_id in run_ids:
		table = run_report(run_id, args[run_id], graph, svg, outdir)
		body += [
			'<tr><td valign="top">',
			'<h3><a href="run{0}.html">Run {0}</a></h3>'.format(run_id),
			'<blockquote>',
			'<table class="table table-striped table-condensed table-mini">',
			table,
			'</td></table></blockquote>',
			'<td align="center" valign="top" style="padding-left:20px;">',
			'<a href="run{0}.html"><img src="run{0}-posteriors.svg" width="400" height="400"/></a>'.format(run_id),
			'<h5>Posterior Distribution</h5></td>',
			'<td align="center" valign="top" style="padding-left:20px;">',
			'<a href="run{0}.html"><img src="run{0}-majority.svg" width="400" height="400"/></a>'.format(run_id),
			'<h5>Majority-Rule Consensus [<a href="run%d-majority.fa">FASTA</a>]</h5></td>' % run_id,
			'</tr>']

	if len(run_ids) > 1:
		# Combine values from multiple runs.
		combined = {
			'nsamples': sum(args[k]['nsamples'] for k in args),
			'time': sum(args[k]['time'] for k in args),
			'mintime': min(args[k]['mintime'] for k in args),
			'maxtime': max(args[k]['maxtime'] for k in args),
			'accept': sum(args[k]['accept'] for k in args),
			'nratios': sum(args[k]['nsamples'] / int(args[k]['thinning']) for k in args)}
		combined['accept_per'] = 100.0 * combined['accept'] / combined['nratios']
		frequencies = {}
		for k in args:
			for e, f in args[k]['frequencies'].iteritems():
				frequencies[e] = frequencies.get(e, 0) + f
		table = [
			('# Samples', '{:,.0f}'.format(combined['nsamples'])),
			('# Accepted', '{:,} ({:.1f}%)'.format(
				combined['accept'], combined['accept_per'])),
			('Total Sample Time', '{:,.1f} sec'.format(combined['time'])),
			('Min. Sample Time', '%.1f sec' % combined['mintime']),
			('Max. Sample Time', '%.1f sec' % combined['maxtime']),
			('Mean Sample Time', '%.1f sec' % (combined['time'] / combined['nsamples']))]
		row = '<tr><td><b>{}</b></td><td>{}</td></tr>'
		table = '\n'.join(row.format(k, v) for k, v in table)
		prefix = os.path.join(outdir, 'run-all')
		posterior_plot(frequencies.copy(), graph, prefix)
		majority_rule_plot(frequencies, graph, 0.5, combined['accept'], prefix)
		body += [
			'<tr><td valign="top">',
			'<h3>All Runs</h3>',
			'<blockquote>',
			'<table class="table table-striped table-condensed table-mini">',
			table,
			'</td></table></blockquote>',
			'<td align="center" valign="top" style="padding-left:20px;">',
			'<img src="run-all-posteriors.svg" width="400" height="400"/>',
			'<h5>Posterior Distribution</h5></td>',
			'<td align="center" valign="top" style="padding-left:20px;">',
			'<img src="run-all-majority.svg" width="400" height="400"/>',
			'<h5>Majority-Rule Consensus [<a href="run-all-majority.fa">FASTA</a>]</h5></td>',
			'</tr>',
			'</table>']

		convergence_plot(
			[arg['trace'] for arg in args.itervalues()],
			os.path.join(outdir, 'convergence.png'))
		body += [
			'<h3>Convergence</h3>',
			'<img src="convergence.png" width="800" height="600"/>',
			'<p><b>(a)</b> Mixing behvaior',
			'<br/><b>(b)</b> Cumulative frequencies of individual edges in each chain',
			'<br/><b>(c)</b> Average st. dev. of the cumulative edge frequencies across chains',
			'<br/><b>(d)</b> Split frequencies at the final sample</p>']

	else:
		body.append('</table>')

	# Final output.
	report.copy_css(outdir)
	report.copy_js(outdir)
	html = html_template % (title, '\n'.join(body))
	open(os.path.join(outdir, 'index.html'), 'w').write(html)


if __name__ == '__main__':
	parser = argparse.ArgumentParser(description="""
		Generates an HTML report of all runs in the agalma diagnostics database
		for the given agalma catalog ID (by default), or of only the specific
		list of RUN_IDS""")
	parser.add_argument('--outdir', '-o', default='./', type=utils.safe_mkdir,
		help="write HTML output to OUTDIR [default: ./]")
	parser.add_argument('run_ids', type=int,  metavar='RUN_ID', nargs='+',
		help="a list of 'sample' pipeline run IDs")
	args = parser.parse_args()
	# Check for graphviz
	if utils.which('dot') is None:
		utils.die("could not find 'dot' command from graphviz")
	index_report(sorted(args.run_ids), args.outdir)

