#
# GABI - Genome Assembly by Bayesian Inference
#
# Copyright 2013-2014, Brown University, Providence, RI. All Rights Reserved.
#
# This file is part of GABI.
#
# GABI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# GABI is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with GABI.  If not, see <http://www.gnu.org/licenses/>.

import math
import os
import scipy.stats
import subprocess

from Bio import SeqIO
from glob import glob
from random import random

from biolite import utils
from biolite.config import get_resource


class Sample:

	def __init__(self, likelihood, prior, length, ncontigs):
		self.posterior = likelihood + prior
		self.likelihood = likelihood
		self.prior = prior
		self.length = length
		self.ncontigs = ncontigs

	def odds_ratio(self, current):
		"""
		"""
		return (not math.isinf(self.likelihood)) and \
		       ((self.posterior - current.posterior) > math.log(random()))


class Sampler:

	def __init__(
		self, reads, length, ncontigs, scale, insert_mean, insert_stdev,
		nodata=False, flatprior=False):

		if nodata:
			self.reads = None
		else:
			self.reads = reads

		if flatprior:
			self.prior = (lambda length, ncontigs : 1.0)
		else:
			# Create a prior distribution from the estimated genome size
			# and number of contigs.
			scale = float(scale)
			length_pdf = scipy.stats.gamma(scale, scale=length/scale).pdf
			ncontigs_pdf = scipy.stats.gamma(scale, scale=ncontigs/scale).pdf
			self.prior = (lambda length, ncontigs :
				math.log(length_pdf(length) * ncontigs_pdf(ncontigs)))

		self.threads = get_resource('threads')
		self.insert_mean = insert_mean
		self.insert_stdev = insert_stdev
	
	def likelihood(self, filename):
		"""
		Estimate assembly likelihood with LAP.
		"""
		if self.reads:
			with open('lap.log', 'w') as f:
				try:
					lap = subprocess.check_output(
						'calc_prob.py -a {} -1 {} -2 {} -o fr -m {} -t {} -X {} -I {} -p {} | sum_prob.py -t 1e-300'.format(
							filename, self.reads[0], self.reads[1],
							int(self.insert_mean), int(self.insert_stdev),
							int(self.insert_mean + 2*self.insert_stdev),
							int(self.insert_mean - 2*self.insert_stdev),
							self.threads),
						shell=True, stderr=f)
					# Clean-up after LAP
					map(os.remove, glob('indexes/temp*'))
				except subprocess.CalledProcessError as e:
					utils.die("failed to call LAP: check lap.log\n", e)
			return float(lap.split()[0])
		else:
			return 1.0

	def sample(self, contigs, filename):
		"""
		Run CGAL wrapper script to calculate log-likelihood of the contigs
		given the reads.
		"""
		# Write out contigs to FASTA file.
		length = 0
		ncontigs = 0
		with open(filename, 'w') as f:
			for contig in contigs:
				SeqIO.write(contig, f, 'fasta')
				length += len(contig.seq)
				ncontigs += 1
		return Sample(
			self.likelihood(filename), self.prior(length, ncontigs),
			length, ncontigs)

